/*
This demo example starts DDC1 streaming (or DDC1 stream playback), shows IF, DDC1, DDC2 and Audio spectrums 
and plays the final audio through speakers.
You can use + and - keys to tune the DDC1.
You can use * and / keys to tune the DDC2.
*/


#include <stdio.h>
#include <dlfcn.h>
#include <errno.h>
#include <unistd.h>
#include <math.h>
#include <pthread.h>
#include <complex.h>
#include <string.h>
#include <curses.h>			// requires curses library
#include "g33ddcapi.h"
#include <pulse/simple.h>	// requires pulse library
#include <pulse/error.h>
#include <time.h>


//#define PLAYBACK		// use this define to test DDC1 playback. Otherwise the receiver DDC1 streaming is used.
//#define DEMO			// use this define to open a demo receiver. Otherwise a real device will be opened.

pa_simple *s = NULL;

typedef double complex cplx;
void CreateWindow();
void fft(cplx buf[],bool copyback);
void Translate(cplx fftbuff[],char *outbuff,bool invert,int *min,int *max);

unsigned long long int TotalAudio=0,TotalDDC1=0,TotalDDC2=0;
unsigned long long int StartTime;

#define FFT_SIZE 64
cplx IF_FFT_Buff[FFT_SIZE];
cplx DDC1_FFT_Buff[FFT_SIZE];
cplx DDC2_FFT_Buff[FFT_SIZE];
cplx Audio_FFT_Buff[FFT_SIZE];

#define DB_PRECISION 8
char dB_Char[DB_PRECISION]={'0','4','5','1','2','6','3','7'};

char BufferIndexArray[4]={'/','-','\\','|'};
int IF_BufferIndex=0,DDC1_BufferIndex=0,DDC2_BufferIndex=0,Audio_BufferIndex=0;

char IF_FFT_Out[6*FFT_SIZE+1];
char DDC1_FFT_Out[6*FFT_SIZE+1];
char DDC2_FFT_Out[6*FFT_SIZE+1];
char Audio_FFT_Out[6*FFT_SIZE+1];

int MinD1=0,MaxD1=0,MinD2=0,MaxD2=0,MinA=0,MaxA=0,MinI=0,MaxI=0;

uint32_t ADC=0,TCXO=0;

class TCallbacks:public IG33DDCDeviceCallback {
public:
    virtual void G33DDC_IFCallback(IG33DDCDevice *Device,const int16_t *Buffer,uint32_t NumberOfSamples,uint16_t MaxADCAmplitude,uint32_t ADCSampleRate) {
		ADC=MaxADCAmplitude;TCXO=ADCSampleRate;
		IF_BufferIndex=(IF_BufferIndex+1)&3;
		for (int i=0;i<FFT_SIZE;i++) IF_FFT_Buff[i]=(double)Buffer[i]/32768.0;
		fft(IF_FFT_Buff,false);
		Translate(IF_FFT_Buff,IF_FFT_Out,false,&MinI,&MaxI);
		}
    virtual void G33DDC_DDC1StreamCallback(IG33DDCDevice *Device,const void *Buffer,uint32_t NumberOfSamples,uint32_t BitsPerSample) {
		DDC1_BufferIndex=(DDC1_BufferIndex+1)&3;
		TotalDDC1+=NumberOfSamples;
		for (int i=0;i<FFT_SIZE;i++) DDC1_FFT_Buff[i]=(double)((int*)Buffer)[i]/2e9+I*((double)((int*)Buffer)[i+1]/2e9);
		fft(DDC1_FFT_Buff,true);
		Translate(DDC1_FFT_Buff,DDC1_FFT_Out,true,&MinD1,&MaxD1);
		}
    virtual int G33DDC_DDC1PlaybackStreamCallback(IG33DDCDevice *Device,void *Buffer,uint32_t NumberOfSamples,uint32_t BitsPerSample) {
#ifdef PLAYBACK
		int *buff=(int*)Buffer;
		for (unsigned int i=0;i<NumberOfSamples;i++) {
			double angle=2*M_PI*10*i/NumberOfSamples;
			buff[i*2]=(int)(2e7*cos(angle));
			buff[i*2+1]=(int)(-2e7*sin(angle));
			}
#endif
		return 1;
		}
    virtual void G33DDC_DDC2StreamCallback(IG33DDCDevice *Device,uint32_t Channel,const float *Buffer,uint32_t NumberOfSamples) {
		DDC2_BufferIndex=(DDC2_BufferIndex+1)&3;
		TotalDDC2+=NumberOfSamples;
		for (int i=0;i<FFT_SIZE;i++) DDC2_FFT_Buff[i]=Buffer[i]+I*(Buffer[i+1]);
		fft(DDC2_FFT_Buff,true);
		Translate(DDC2_FFT_Buff,DDC2_FFT_Out,true,&MinD2,&MaxD2);
		}
    virtual void G33DDC_DDC2PreprocessedStreamCallback(IG33DDCDevice *Device,uint32_t Channel,const float *Buffer,uint32_t NumberOfSamples,float SlevelPeak,float SlevelRMS) {}
    virtual void G33DDC_AudioStreamCallback(IG33DDCDevice *Device,uint32_t Channel,uint32_t Type,const float *Buffer,uint32_t NumberOfSamples) {
		if (!Channel&&!Type) {		// first demodulator (RX) and 
			TotalAudio+=NumberOfSamples;
			Audio_BufferIndex=(Audio_BufferIndex+1)&3;
			pa_simple_write(s,Buffer,(size_t)NumberOfSamples*sizeof(float),NULL);
			for (int i=0;i<FFT_SIZE;i++) Audio_FFT_Buff[i]=Buffer[i];//+I*(Buffer[i+1]);
			fft(Audio_FFT_Buff,false);
			Translate(Audio_FFT_Buff,Audio_FFT_Out,false,&MinA,&MaxA);
			}
		}
    virtual int G33DDC_AudioPlaybackStreamCallback(IG33DDCDevice *Device,uint32_t Channel,float *Buffer,uint32_t NumberOfSamples) {return 0;}
	};

int main(void)
{
G33DDCAPI_CREATE_INSTANCE CreateInstance;
void *API;
IG33DDCDevice *Device;
TCallbacks callbacks;
CreateWindow();
int Freq1=702000,Freq2=0;
int error=0;
struct timespec tt;

pa_sample_spec ss;
ss.format=PA_SAMPLE_FLOAT32LE,
ss.rate=48000;
ss.channels=1;

if (!(s=pa_simple_new(NULL,"g33ddcdemo",PA_STREAM_PLAYBACK,NULL,"playback",&ss,NULL,NULL,&error))) fprintf(stderr, __FILE__": pa_simple_new() failed: %s\n", pa_strerror(error));

for (int i=0;i<FFT_SIZE;i++) {
	IF_FFT_Out[6*i]=DDC1_FFT_Out[6*i]=DDC2_FFT_Out[6*i]=Audio_FFT_Out[6*i]=27;
	IF_FFT_Out[6*i+1]=DDC1_FFT_Out[6*i+1]=DDC2_FFT_Out[6*i+1]=Audio_FFT_Out[6*i+1]='[';
	IF_FFT_Out[6*i+2]=DDC1_FFT_Out[6*i+2]=DDC2_FFT_Out[6*i+2]=Audio_FFT_Out[6*i+2]='4';
	IF_FFT_Out[6*i+3]=DDC1_FFT_Out[6*i+3]=DDC2_FFT_Out[6*i+3]=Audio_FFT_Out[6*i+3]='0';
	IF_FFT_Out[6*i+4]=DDC1_FFT_Out[6*i+4]=DDC2_FFT_Out[6*i+4]=Audio_FFT_Out[6*i+4]='m';	
	IF_FFT_Out[6*i+5]=DDC1_FFT_Out[6*i+5]=DDC2_FFT_Out[6*i+5]=Audio_FFT_Out[6*i+5]=' ';	
	}
IF_FFT_Out[6*FFT_SIZE/2]=Audio_FFT_Out[6*FFT_SIZE/2]=0;
DDC1_FFT_Out[6*FFT_SIZE]=DDC2_FFT_Out[6*FFT_SIZE]=0;

API=dlopen("libg33ddcapi.so",RTLD_LAZY);

if (API) {
	CreateInstance=(G33DDCAPI_CREATE_INSTANCE)dlsym(API,"CreateInstance");
	if(CreateInstance(G33DDC_CLASS_ID_DEVICE,(void**)&Device)) {
#ifdef DEMO
		if(Device->Open(G3XDDC_OPEN_DEMO)) {
#else
		if(Device->Open(G3XDDC_OPEN_FIRST)) {
#endif
			G33DDC_DEVICE_INFO info;
			G3XDDC_DDC_INFO info1={0},info2={0};
			Device->GetDeviceInfo(&info,sizeof(info));
			printf("Device %8s opened, flags=#%x\n",info.SerialNumber,info.Flags);

			Device->SetCallback(&callbacks);
			int ret=Device->SetPower(1); printf("SetPower, ret=%d, errno=%d\n",ret,errno);
			ret=Device->StartIF(1000); printf("StartIF, ret=%d, errno=%d\n",ret,errno);
			ret=Device->SetDDC1(12); printf("SetDDC1, ret=%d, errno=%d\n",ret,errno);
			Device->GetDDC1(NULL,&info1);
#ifdef PLAYBACK
			ret=Device->StartDDC1Playback(30000,32); printf("StartDDC1Playback, ret=%d, errno=%d\n",ret,errno);
#else
			ret=Device->StartDDC1(50000); printf("StartDDC1, ret=%d, errno=%d\n",ret,errno);
			ret=Device->SetDDC1Frequency(Freq1); printf("SetDDC1Freq, ret=%d, errno=%d\n",ret,errno);
#endif			
			clock_gettime(CLOCK_REALTIME,&tt);
			StartTime=tt.tv_sec;
			StartTime*=1000000000;
			StartTime+=tt.tv_nsec;

			ret=Device->StartDDC2(0,40000); printf("StartDDC2, ret=%d, errno=%d\n",ret,errno);
			Device->GetDDC2(NULL,&info2);
			ret=Device->SetDDC2Frequency(0,Freq2); printf("SetDDC2Freq, ret=%d, errno=%d\n",ret,errno);
			Device->SetDemodulatorMode(0,G3XDDC_MODE_AM);
			ret=Device->StartAudio(0,2000); printf("StartAudio, ret=%d, errno=%d\n",ret,errno);
			Device->SetDemodulatorFilterBandwidth(0,6000);
			Device->SetAGC(0,TRUE);
			Device->SetAGCParams(0,0.1,0.1,-10);
/*
//*/
			puts("Streaming started, press enter");
			getchar();
			printf("\E[H\E[2J");
			WINDOW *w=initscr();
			cbreak();
			noecho();
			nodelay(w,TRUE);
			
			bool end=false;
			do {
				printf("\033[2;0HDDC1: %dkHz/%dkHz, Freq=%dkHz    \r\nDDC2=%dkHz/%dkHz, Freq=%dkHz  \r\nADC=%u, TCXO=%u    \r\n",info1.SampleRate/1000,info1.Bandwidth/1000,Freq1/1000,info2.SampleRate/1000,info2.Bandwidth/1000,Freq2/1000,ADC,TCXO);
				printf("\033[8;0HIF:     %s\x1b[0m%c %d+%d  ",IF_FFT_Out,(unsigned int)BufferIndexArray[IF_BufferIndex],MinI,MaxI);
				printf("\033[10;0HDDC1: %s\x1b[0m%c %d+%d  ",DDC1_FFT_Out,BufferIndexArray[DDC1_BufferIndex],MinD1,MaxD1);
				printf("\033[12;0HDDC2: %s\x1b[0m%c %d+%d  ",DDC2_FFT_Out,BufferIndexArray[DDC2_BufferIndex],MinD2,MaxD2);
				printf("\033[14;0HAudio: %s\x1b[0m%c %d+%d  ",Audio_FFT_Out,BufferIndexArray[Audio_BufferIndex],MinA,MaxA);
				usleep(10000);
				int c=getch();
				if (c==27) end=true;
				else if (c=='+') {Freq1+=1000;Device->SetDDC1Frequency(Freq1);}
				else if (c=='-') {Freq1-=1000;Device->SetDDC1Frequency(Freq1);}
				else if (c=='/') {Freq2+=1000;Device->SetDDC2Frequency(0,Freq2);}
				else if (c=='*') {Freq2-=1000;Device->SetDDC2Frequency(0,Freq2);}
				} while (!end);
			clock_gettime(CLOCK_REALTIME,&tt);
			unsigned long long int t=tt.tv_sec;
			t*=1000000000;
			t+=tt.tv_nsec;
			t-=StartTime;
			t/=1000000;

			endwin();

			printf("DDC1=%d kHz, DDC2=%dkHz, Audio=%d kHz\n\n",(unsigned int)((TotalDDC1+t/2)/t),(unsigned int)((TotalDDC2+t/2)/t),(unsigned int)((TotalAudio+t/2)/t));

			puts("Stopping IF");
			Device->StopIF();
			puts("Stopping Audio");
			Device->StopAudio(0);
			puts("Stopping DDC2");
			Device->StopDDC2(0);
			puts("Stopping DDC1");
			Device->StopDDC1();
			puts("Closing the device");
			Device->Close();
			puts("Device closed");		
			}
		else printf("Failed to open device. Error code=%d\n",errno);
		puts("Releasing the device");
		Device->Release();
		}
	else printf("Failed to create device object. Error code=%d\n",errno);
	puts("Closing the API");
	dlclose(API);
    }
else printf("Failed to load libg33ddcapi.so. %s\n",dlerror());

if (s) pa_simple_free(s);
puts("Closing the demo");
return 0;
}

//---------------------------------------------------------------------------------------------------
double window[FFT_SIZE];
void CreateWindow()
{
double sum=0.0;
for (int i=0;i<FFT_SIZE;i++) sum+=(window[i]=0.42-0.5*cos(2.0*M_PI*i/double(FFT_SIZE-1))+0.08*cos(4.0*M_PI*i/double(FFT_SIZE-1)));
sum=1.0/sum;
for (int i=0;i<FFT_SIZE;i++) window[i]*=sum;
}
//--------------------------------------------------------------------------------
void Translate(cplx fftbuff[],char *outbuff,bool invert,int *min,int *max)
{
double maxdB=-100000,mindB=0;
for (int i=0;i<FFT_SIZE;i++) {
	double a=10*log10(creal(fftbuff[i])*creal(fftbuff[i])+cimag(fftbuff[i])*cimag(fftbuff[i])+1e-20);
	fftbuff[i]=a;
	if (a>maxdB) maxdB=a;
	if (a<mindB) mindB=a;
	}
maxdB-=mindB;
if (invert) for (int i=0;i<FFT_SIZE/2;i++) {
	unsigned int index=(unsigned int)((DB_PRECISION-1)*((creal(fftbuff[i])-mindB)/maxdB));
	outbuff[6*(FFT_SIZE/2-1-i)+3]=dB_Char[index];
	index=(unsigned int)((DB_PRECISION-1)*((creal(fftbuff[FFT_SIZE/2+i])-mindB)/maxdB));
	outbuff[6*(FFT_SIZE-1-i)+3]=dB_Char[index];
	}
else for (int i=0;i<FFT_SIZE;i++) {
	unsigned int index=(unsigned int)((DB_PRECISION-1)*((creal(fftbuff[i])-mindB)/maxdB));
	if (index>=DB_PRECISION) index=DB_PRECISION-1;
	outbuff[6*i+3]=dB_Char[index];
	}
*min=(int)mindB;
*max=(int)maxdB;
}
//--------------------------------------------------------------------------------
void _fft(cplx buf[], cplx out[],int step)
{
	if (step<FFT_SIZE) {
		_fft(out,buf,step*2);
		_fft(out+step,buf+step,step*2);
 
		for (int i=0;i<FFT_SIZE;i+=2*step) {
			cplx t=cexp(-1*I*M_PI*i/FFT_SIZE)*out[i+step];
			buf[i/2]=out[i]+t;
			buf[(i+FFT_SIZE)/2]=out[i]-t;
		}
	}
}
void fft(cplx buf[],bool copyback)
{
	cplx out[FFT_SIZE];
	for (int i=0;i<FFT_SIZE;i++) out[i] = buf[i]*window[i];
	_fft(buf,out,1);
if (copyback) for (int i=0;i<FFT_SIZE;i++) buf[i]=out[i];
}
//--------------------------------------------------------------------------------
